
import os

import logging

import lerc_api

logger = logging.getLogger("lerc_control."+__name__)

def delete_file(client, file_path):
    return False

def delete_registry(client, reg_path):
    return False

def delete_service(client, service_name):
    return False

def delete_scheduled_task(client, task_name):
    return False

def delete_directory(client, dir_path):
    return False

def kill_process(client, pname=None, pid=None):
    if pname is None and pid is None:
        logger.error("Must supply a process name or a process id to kill.")
        return False
    return False

def remediate(client, remediation_script):
    if not os.path.exists(remediation_script):
        logger.error("'{}' Does not exist".format(remediation_script))
        return False

    





import os
import sys
import time
import json
import logging
import requests
from hashlib import md5
from datetime import datetime
from contextlib import closing
from configparser import ConfigParser


def check_config(config, required_keys):
    """Just make sure the keys show up somewhere in the config - Not a fool-proof check

    :param ConfigParser config: A configparser.ConfigParser
    :param list required_keys: A list of required config keys
    :return: False if a key is missing, else config
    """
    logger = logging.getLogger(__name__+".check_config")
    at_least_one_missing = False
    if isinstance(required_keys, list) and required_keys:
        for key in required_keys:
            found = False
            for section in config.sections():
                if config.has_option(section, key):
                    found = True
                    break
            if not found:
                at_least_one_missing = True
                logger.error("Missing required config item: {}".format(key))
    if at_least_one_missing:
        return False
    return config


def load_config(profile='default', required_keys=[]):
    """Load lerc configuration. Configuration files are looked for in the following locations::
        /<python-lib-where-lerc_control-installed>/etc/lerc.ini
        /etc/lerc_control/lerc.ini
        /opt/lerc/lerc_control/etc/lerc.ini
        ~/<current-user>/.lerc_control/lerc.ini

    Configuration items found in later config files take presendence over earlier ones.

    :param str profile: (optional) Specifiy a group or company to work with.
    :param list required_keys: (optional) A list of required config keys to check for
    """
    logger = logging.getLogger(__name__+".load_config")
    config = ConfigParser()
    config_paths = []
    # Get the working lerc_control directory
    BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    # default
    config_paths.append(os.path.join(BASE_DIR, 'etc', 'lerc.ini'))
    # global
    config_paths.append('/etc/lerc_control/lerc.ini')
    # legacy
    config_paths.append('/opt/lerc/lerc_control/etc/lerc.ini')
    # user specific
    config_paths.append(os.path.join(os.path.expanduser("~"),'.lerc_control','lerc.ini'))
    finds = []
    for cp in config_paths:
        if os.path.exists(cp):
            logger.debug("Found config file at {}.".format(cp))
            finds.append(cp)
    if not finds:
        logger.critical("Didn't find any config files defined at these paths: {}".format(config_paths))
        raise Exception("MissingLercConfig", "Config paths : {}".format(config_paths))

    config.read(finds)
    try:
        config[profile]
    except KeyError:
        logger.critical("No section named '{}' in configuration files : {}".format(profile, config_paths))
        raise

    if required_keys:
        check_config(config, required_keys)

    return config


######################
## lerc_api GLOBALS ##
######################
CONFIG = load_config()
QUERY_FIELDS = ['cmd_id', 'hostname', 'operation', 'cmd_status', 'client_id', 'client_status', 'version', 'company_id', 'company']
QUERY_FIELD_DICT = {'cmd_id': 'The ID of a specific command',
                    'hostname': 'The hostname of a client',
                    'operation': 'A Command operation type: upload,run,download,quit',
                    'cmd_status': 'The status of a command: pending,preparing,complete,error,unknown,started',
                    'client_id': 'Specify a LERC by ID',
                    'client_status': 'A LERC status: busy,online,offline,unknown,uninstalled',
                    'version': 'The LERC version string',
                    'company_id': 'Specify a company/group ID',
                    'company': 'A company/group name'}
QUERY_FIELD_DESCRIPTIONS = []
for key,value in QUERY_FIELD_DICT.items():
    QUERY_FIELD_DESCRIPTIONS.append({'field': key, 'description': QUERY_FIELD_DICT[key]})

def parse_lerc_server_query(query_str):
    """This function converts a string from field:value pairs into \*\*args that lerc_session.query can recognize.

    :param str query_str: A query string to be parsed.
    :return: \*\*args ready for lerc_session.Query()
    """
    logger = logging.getLogger(__name__+".parse_lerc_server_query")
    query_parts = query_str.split()
    args = {}
    negated_with_not = False
    for part in query_parts:
        negated = False
        if part[0] == '-' or part[0] == '!':
            negated = True
            part = part[1:]
        elif part.upper() == 'NOT':
            negated_with_not = True
            continue
        if negated_with_not:
            negated = True
            negated_with_not = False
        field = part[:part.find(':')]
        if field not in QUERY_FIELDS:
            logger.critical("{} is not a valid query filed. Valid fields: {}".format(field, QUERY_FIELDS))
            return False
        value = part[part.find(':')+1:]
        if negated:
            value = '-' + value
        args[field] = value
    return args


class Client():
    """Represents a Live Endpoint Response Client and is used to interact with this LERC.

    :param dict host_dict: A dictionary representaion of a LERC
    """

    contained = False
    logger = logging.getLogger(__name__+".Client")

    def __init__(self, host_dict, profile='default'):
        self.hostname = host_dict['hostname']
        self.version = host_dict['version']
        self.company_id = host_dict['company_id']
        self.id = host_dict['id']
        self.status = host_dict['status']
        self.sleep_cycle = host_dict['sleep_cycle']
        self.last_activity = host_dict['last_activity']
        self.install_date = host_dict['install_date']
        self._raw = host_dict
        self.profile = profile
        self._lerc_session = lerc_session(profile=profile)

    @property
    def get_dict(self):
        """Return the client as it's dictionary representation."""
        return self._raw

    def refresh(self):
        """Query the LERC Server to get an update on this LERC."""
        r = requests.get(self._ls.server+'/query', params={'client_id': self.id}, cert=self._ls.cert).json()
        if 'error' in r:
            self.logger.error("{}".format(r['error']))
            return False
        if 'clients' in r and len(r['clients']) == 1:
            r = r['clients'][0]
        else:
            self.logger.error("Unexpected result from server : {}".format(r))
            return False
        self.status = r['status']
        self.last_activity = r['last_activity']
        self.sleep_cycle = r['sleep_cycle']
        self.version = r['version']
        self.install_date = r['install_date']
        self._raw = r
        return True

    @property
    def is_online(self):
        """Return True if the client is Online."""
        self.refresh() 
        if self.status == 'ONLINE':
            return True
        return False

    @property
    def is_busy(self):
        """Return True is the client is Busy working on a command."""
        self.refresh()
        if self.status == 'BUSY':
            return True
        return False

    @property
    def _ls(self):
        """The Client's lerc_session object."""
        return self._lerc_session

    def __str__(self):
        text = "\n"
        text += "\tClient ID: {}\n".format(self.id)
        text += "\tHostname: {}\n".format(self.hostname)
        text += "\tStatus: {}\n".format(self.status)
        text += "\tCompany ID: {}\n".format(self.company_id)
        config = self._ls.get_config
        for section in config.sections():
            if config.has_option(section, 'company_id'):
                if config[section].getint('company_id') == int(self.company_id):
                    text += "\tCompany Name: {}\n".format(section)
        text += "\tVersion: {}\n".format(self.version)
        text += "\tSleep Cycle: {} seconds\n".format(self.sleep_cycle)
        text += "\tInstall Date: {}\n".format(self.install_date)
        text += "\tLast Activity: {}\n\n".format(self.last_activity)
        return text

    def _issue_command(self, command):
        # check the host, make the post
        arguments = {'host': self.hostname, 'id': self.id}
        headers={"content-type": "application/json"}
        r = requests.post(self._ls.server+'/command', cert=self._ls.cert, headers=headers,
                                                params=arguments, data=json.dumps(command))
        if r.status_code != requests.codes.ok:
            self.error = { 'status_code': r.status_code, 'message': "ERROR : {}".format(r.text) }
            self.logger.error(self.error['message'])
            return False
        else: # record the command
            result = r.json()
            if 'command' in result:
                self.logger.info("{} (CID: {})".format(result['message'], result['command']['command_id']))
                return Command(result['command'], profile=self.profile)
            else:
                if 'error' not in result:
                    raise Exception("Unexpected result: {}".format(result))
                self.logger.error("{}".format(results['error']))
                return False


    def Run(self, shell_command, async=False):
        """Execute a shell command on the host.

        :param str shell_command: The command to run on the host
        :param bool async: (optional) ``False``: LERC client will stream any results and  wait until for completion. ``True``: Execute the command and return immediately.
        """
        command = { "operation":"run", "command": shell_command, "async": async }
        return self._issue_command(command)

    def Download(self, server_file_path, client_file_path=None, analyst_file_path=None):
        """Instruct a client to download a file. The file will be streamed to the server if the server doesn't have it yet. The streamed to the client.

        :param str server_file_path: The path to the file that you want the client to download. If the supplied argument is the path to the file on the system this function is called, an attempt will be made to complete the analyst_file_path automatically.
        :param str client_file_path: (optional) where the client should write the file, defaults server config default directory + the file name taken off of the server_file_path.
        :param str analyst_file_path: (optional) path to the original file the analyst - This allows an analyst to resume a transfer between a lerc_session and the server. Neccessary for streaming the file to the server, if the server does not already have it
        """
        if analyst_file_path is None:
            analyst_file_path = server_file_path

        if '/' in server_file_path:
            server_file_path = server_file_path[server_file_path.rfind('/')+1:]

        command = { "operation":"download", "server_file_path":server_file_path,
                    "client_file_path":client_file_path, "analyst_file_path":analyst_file_path}

        command = self._issue_command(command)
        command.stream_file(analyst_file_path)
        return command

    def Upload(self, path):
        """Upload a file from client to server. The file will be streamed to the server and then streamed to the lerc_session.

        :param str path: The path to the file, on the endpoint
        """
        command = { "operation":"upload", "client_file_path":path }
        return self._issue_command(command)

    def Quit(self):
        """The Quit command tells the LERC client to uninstall itself from the endpoint.
        """
        command = { "operation":"quit" }
        return self._issue_command(command)

    def list_directory(self, dir_path):
        """List the given directory, read the results into a list and return the list.
        """
        command = self.Run('cd "{}" && dir'.format(dir_path))
        command.wait_for_completion()
        results = command.get_results(return_content=True)
        dir_lines = results.decode('utf-8').splitlines()
        results = {}
        results['dir_lines'] = dir_lines
        contents = []
        for line in dir_lines:
            content = {}
            if line:
                if line[2] == '/':
                    content['name'] = line[line.rfind(' ')+1:]
                    if 'DIR' in line:
                        content['type'] = 'DIR'
                        content['size'] = ''
                    else:
                        content['type'] = 'file'
                        line_no_name = line[:line.rfind(' ')]
                        content['size'] = line_no_name[line_no_name.rfind(' ')+1:]
                    contents.append(content)
                elif 'bytes' in line:
                    if 'File' in line:
                        line = line[:line.rfind(' ')]
                        dir_size = line[line.rfind(' ')+1:]
                        results['dir_size'] = dir_size
                    elif 'Dir' in line:
                        line = line[:line.rfind(' ')]
                        line = line[:line.rfind(' ')]
                        drive_free_space = line[line.rfind(' ')+1:]
                        results['drive_free_space'] = drive_free_space 

        results['dir_dict'] = contents
        return results

    def contain(self):
        """Use the windows firewall to isolate a host. Everything will be blocked but lerc's access outbound. You must attach to a host before using contain.
        This is implemented by dropping a bat file (a default is included in the lerc_control/tools dir) that will isolate a host with the windows firewall and only allow the lerc.exe client access through the firewall. The bat file will pause for ~60 seconds without prompting the user. Lerc is issued a run command to kill the bat file. If Lerc is able to fetch this run command from the control server, the bat file will be killed before the 60 seconds are up. If not, the bat file will undo all firewall changes.

        :return: True on success
        """

        if self.contained:
            return self.contained

        self.logger.info("containing host..")
        self.refresh()
        safe_contain_bat_path = self._ls.config[self._ls.profile]['containment_bat']
        contain_cmd = self._ls.config[self._ls.profile]['contain_cmd']

        self.Download(safe_contain_bat_path)
        containment_command = self.Run(contain_cmd.format(int(self.sleep_cycle)+5), async=True)

        # Dummy command to give the containment command enough time to execute before lerc kills it with wmic
        self.Run("dir")

        bat_name = safe_contain_bat_path[safe_contain_bat_path.rfind('/')+1:]
        kill_command = self.Run('wmic process where "CommandLine Like \'%{}%\'" Call Terminate'.format(bat_name))

        # for spot checking
        check_command = self.Run('netsh advfirewall show allprofiles')

        if not containment_command.wait_for_completion():
            self.logger.error("Containment command failed : {}".format(containment_command))
            return False
        self.logger.debug("{}".format(containment_command))
        self.logger.info("Host contained at: {}".format(datetime.now()))

        self.logger.info("Command {} should return before {} seconds have passed.".format(kill_command.id, self.sleep_cycle))
        if not kill_command.wait_for_completion():
            self.logger.error("Waiting for kill command failed: {}".format(kill_command))
            return False
        self.logger.debug("{}".format(kill_command))

        self.logger.info("Getting firewall status for due diligence..")
        check_command.wait_for_completion()
        check_command.get_results(file_path = "{}_{}_firewall_status.txt".format(self.hostname, check_command.id), print_run=False)

        self.contained = True
        return self.contained

    # Move to Client class
    def release_containment(self):
        """Release containment on client.

        :return: True on success.
        """
        reset_cmd = self.Run("netsh advfirewall reset && netsh advfirewall show allprofiles")
        reset_cmd.wait_for_completion()
        self.logger.info("Host containment removed at: {}".format(datetime.now()))
        self.logger.info("Getting firewall status for due diligence..")
        reset_cmd.get_results(file_path = "{}_{}_firewall_reset.txt".format(self.hostname, reset_cmd.id), print_run=False)
        self.contained = False
        return not self.contained




class Command():
    """Represents a Live Endpoint Response Client Command.

    :param dict cmd_dict: A dictionary representation of a LERC Command.
    """

    logger = logging.getLogger(__name__+".Command")

    def __init__(self, cmd_dict, profile='default'):
        self.id = cmd_dict['command_id']
        self.hostname = cmd_dict['hostname']
        self.client_id = cmd_dict['client_id']
        self.operation = cmd_dict['operation']
        self.command = cmd_dict['command']
        self.status = cmd_dict['status']
        self.client_file_path = cmd_dict['client_file_path']
        self.server_file_path = cmd_dict['server_file_path']
        self.analyst_file_path = cmd_dict['analyst_file_path']
        self.async_run = cmd_dict['async_run']
        self.file_position = cmd_dict['file_position']
        self.filesize = cmd_dict['filesize']
        self._lerc_session = lerc_session(profile=profile)
        self._raw = cmd_dict

    def __str__(self):
        text = "\n\t----------------------------\n"
        text += "\tCommand ID: {}\n".format(self.id)
        text += "\tClient ID: {}\n".format(self.client_id)
        text += "\tHostname: {}\n".format(self.hostname)
        text += "\tOperation: {}\n".format(self.operation)
        if self.operation == 'RUN':
            text += "\t  |-> Command: {}\n".format(self.command)
            text += "\t  |-> Async: {}\n".format(self.async_run)
        text += "\tStatus: {}\n".format(self.status)
        text += "\tClient Filepath: {}\n".format(self.client_file_path)
        text += "\tServer Filepath: {}\n".format(self.server_file_path)
        text += "\tAnalyst Filepath: {}\n".format(self.analyst_file_path)
        text += "\tFile Position: {}\n".format(self.file_position)
        text += "\tFile Size: {}\n\n".format(self.filesize)
        return text

    @property
    def _ls(self):
        return self._lerc_session

    @property
    def get_dict(self):
        return self._raw

    def refresh(self, cmd_dict=None):
        """Query the lerc server and update this command with the latest data."""
        if not cmd_dict:
            r = requests.get(self._ls.server+'/query', cert=self._ls.cert, params={'cmd_id': self.id, 'rc': True}).json()
            if 'commands' in r and len(r['commands']) == 1:
                cmd_dict = r['commands'][0]
            else:
                self.logger.warning("No command by id:{}".format(self.id))
                return False
        self.status = cmd_dict['status']
        self.file_position = cmd_dict['file_position']
        self.filesize = cmd_dict['filesize']
        self.server_file_path = cmd_dict['server_file_path']

    def stream_file(self, file_path, position=0):
        """Stream a file to the lerc server that the server needs to for this command."""
        # file_path - file to send
        # position - position in file to send from (resume capable)
        if not os.path.exists(file_path):
            self.logger.error("{} does not exists. Aborting.".format(file_path))
            return False

        # get md5 of file
        md5_hasher = md5()
        with open(file_path, 'rb') as f:
            for chunk in iter(lambda: f.read(4096), b''):
                md5_hasher.update(chunk)
        file_md5 = md5_hasher.hexdigest().lower()

        statinfo = os.stat(file_path)
        arguments = {'host': self.hostname, 'cid': self.id, 'filesize': statinfo.st_size, 'md5': file_md5}
        def gen():
            with open(file_path, 'rb') as f:
                f.seek(position)
                data = f.read(4096)
                while data:
                    yield data
                    data = f.read(4096)
        r = requests.post(self._ls.server+'/command/upload', cert=self._ls.cert, params=arguments, data=gen()).json()
        if 'error' in r:
            self.logger.error("{}".format(r['error']))
            return r
        if 'warn' in r:
            self.logger.info(r['warn'])
        self.refresh(cmd_dict=r)
        return True

    def get_error_report(self):
        """If an error report exists for this command, get it.
        """
        arguments = {'cid': self.id, 'error': True}
        r = requests.get(self._ls.server+'/command/download', cert=self._ls.cert, params=arguments).json()
        return r


    def get_results(self, file_path=None, chunk_size=None, print_run=True, return_content=False, position=0):
        """Get any results available for a command. If cid is None, any cid currently assigned to the lerc_session will be used.

        :param int cid: (optional) The Id of a command to work with.
        :param str file_path: (optional) The path to write the results. default: <hostname>_<cid>_filename to current dir.
        :param int chunk_size: (optional) Specify the size of the chunks (bytes) to stream with the server
        :param boolean print_run: (optionl default:True) If True, print Run command results to console
        :param boolean return_content: (content) Do not write the results to a file, return the results as a byte string
        :param int position: (optional) For manually specifing byte position
        :return: If return_content==True, the raw content will be returned as a byte string, else the command is return on success. 
        """
        # make sure the command is up-to-date
        self.refresh()

        if self.operation == 'DOWNLOAD' or self.operation == 'QUIT':
            self.logger.info("No results to get for '{}' operations".format(self.operation))
            return None
        elif self.filesize == 0:
            self.logger.info("Command complete. No output to collect.")
            return None

        # Proceed only if the server reports it has results
        r = requests.get(self._ls.server+'/command/download', cert=self._ls.cert, params={'result': self.server_file_path}).json()
        if 'error' in r:
            self.logger.warning("{}".format(r['error']))
            return False

        if chunk_size:
            self._ls.chunk_size = chunk_size

        if not file_path:
            file_path = self.server_file_path
        file_path = file_path[file_path.rfind('/')+1:] if file_path.startswith('data/') else file_path

        # Do we already have some of the result file?
        if os.path.exists(file_path):
            statinfo = os.stat(file_path)
            position = statinfo.st_size
            self.logger.info("Already have {} out of {} bytes. Resuming download from server.."
                             .format(position, self.filesize))
        else:
            self.logger.debug("getting results for {}".format(self.id))

        arguments = {'position': position, 'cid': self.id}
        headers = {"Accept-Encoding": '0'}

        if self.status != 'COMPLETE' and self.status != 'STARTED':
            self.logger.warning("Any results for commands in state={} can not be reliably streamed.".format(self.status))
            return requests.get(self._ls.server+'/command/download', cert=self._ls.cert, params=arguments).json()

        raw_bytes = None
        total_chunks, remaining_bytes = divmod(self.filesize - position, self._ls.chunk_size)
        with closing(requests.get(self._ls.server+'/command/download', cert=self._ls.cert, params=arguments, headers=headers, stream=True)) as r:
            try:
                r.raw.decode_content = True
                # are we returning the raw bytes?
                if return_content:
                    raw_bytes = b''
                    for i in range(total_chunks):
                        raw_bytes += r.raw.read(self._ls.chunk_size)
                    raw_bytes += r.raw.read(remaining_bytes)
                    return raw_bytes
                else:
                    with open(file_path, 'ba') as f:
                        for i in range(total_chunks):
                            if self.operation == 'RUN' and print_run:
                                print(r.raw.read(self._ls.chunk_size).decode('utf-8'))
                            f.write(r.raw.read(self._ls.chunk_size))
                        final_chunk = r.raw.read(remaining_bytes)
                        if self.operation == 'RUN' and print_run:
                            print(final_chunk.decode('utf-8'))
                        f.write(final_chunk)
                        f.close()
            except Exception as e:
                self.logger.error(str(e))
                return False

        # Did we get the entire result file?
        filesize = os.stat(file_path).st_size
        if self.filesize == filesize:
            self.logger.info("Result file download complete. Wrote {}.".format(file_path))
            return True
        else:
            self.logger.info("Data stream closed prematurely. Have {}/{} bytes. Trying to resume..".
                              format(filesize, self.filesize))
            self.get_results()
        return None

    def wait_for_completion(self):
        """Wait for this command to complete by continously querying for its status to change to 'COMPLETE' with the server.

        :return: True when a command's status changes to 'COMPLETE'. 
        """
        while True:
            self.refresh()
            tmp_client = self._ls.get_client(self.client_id)
            #if tmp_client.status != 'ONLINE' and tmp_client.status != 'BUSY':
            if not tmp_client.is_online and not tmp_client.is_busy:
                self.logger.warning("This command's LERC ({} (ID:{})) has gone to a status of '{}'".format(self.hostname, self.client_id, tmp_client.status))
            if self.status == 'PENDING': # we wait
                self.logger.info("Command {} PENDING. Checking again in 10 seconds..".format(self.id))
                time.sleep(10)
            elif self.status == 'PREPARING': # the command needs something from us (file)
                if self.operation != "DOWNLOAD":
                    self.logger.error("Command {} is PREPARING but we don't know what for.".format(self.id))
                    return False
                self.logger.info("Command {} PREPARING. Streaming file to server..".format(self.id))
                # function to stream file to server
                if not self.analyst_file_path:
                    self.logger.error("Can't resume upload to server, analyst_file_path is not defined")
                    return False
                self.stream_file(self.analyst_file_path, self.file_position)
            elif self.status == 'STARTED':
                self.logger.info("Command {} STARTED. Checking again in 10 seconds..".format(self.id))
                time.sleep(10)
            elif self.status == 'COMPLETE':
                self.logger.info("Command {} COMPLETE.".format(self.id))
                return True
            else: # Only here if command in UNKNOWN or ERROR state
                self.logger.info("Command {} state: {}.".format(self.id, self.status))
                if self.status == 'ERROR':
                    err = self.get_error_report()
                    self.logger.warning("Error message for command={} : {}".format(self.id, err['error']))
                return None



class lerc_session():
    """Represents a Live Endpoint Response Client Server control session.
    This class is for interacting and managing the LERC clients and server.

    Optional arguments:

    :profile: Specifiy a group or company to work with. These are defined in the lerc.ini config file.
    :server: The name the LERC control server to work with. Default is read from the lerc.ini config file.
    :chunk_size: The chunk size to use when streaming files between a lerc_session and the LERC server

    """
    server = None
    logger = logging.getLogger(__name__+".lerc_session")

    def __init__(self, profile='default', server=None, chunk_size=4096):
        self.config = check_config(CONFIG, required_keys=['server', 'server_ca_cert', 'client_cert', 'client_key'])
        self.profile = profile
        if server:
            self.server = server
        else:
            self.server = self.config[profile]['server']
        if not self.server.startswith('https://'):
            self.server = 'https://' + self.server
        if self.server[-1] == '/':
            self.server = self.server[:-1]
        if 'server_ca_cert' in self.config[profile]:
            self.logger.debug("setting 'REQUESTS_CA_BUNDLE' environment variable for HTTPS verification")
            os.environ['REQUESTS_CA_BUNDLE'] = self.config[profile]['server_ca_cert']
        self.client_cert = self.config[profile]['client_cert']
        self.client_key = self.config[profile]['client_key']
        self.cert = (self.client_cert, self.client_key)
        if 'ignore_system_proxy' in self.config[profile]:
            if self.config[profile].getboolean('ignore_system_proxy'):
                # route direct
                if 'https_proxy' in os.environ:
                    del os.environ['https_proxy']
        self.chunk_size = chunk_size

    @property
    def get_config(self):
        return self.config

    @property
    def get_profile(self):
        return self.profile

    def query(self, **kwargs):
        """Query the lerc server database. Very simple queries supported.
        
        Note: The 'rc' field stands for return commands. By default, command results are only returned if 'rc' is set OR if a 'cmd*' field is set.

        Possible kwargs:

        :param str cmd_id: Command id
        :param str hostname: lerc hostname
        :param str operation: Lerc operation
        :param str cmd_status: Status of a command
        :param str client_id: ID of a lerc
        :param str client_status: Status of a lerc
        :param str version: lerc version
        :param str company_id: ID of a company/group that lercs are in
        :param str company: Name of a company/group that lercs are in
        """
        valid_fields = ['rc', 'cmd_id', 'hostname', 'operation', 'cmd_status', 'client_id', 'client_status', 'version', 'company_id', 'company']
        #valid_fields = QUERY_FIELDS.append('rc')
        keys = kwargs.keys()
        valid_keys = [key for key in keys if key in valid_fields]
        if len(valid_keys) <= 0:
            self.logger.warning("None of {} are valid query fileds.".format(keys))
            return False

        # If cmd keys were passed, assume it's ok to set the return commands arg
        if 'rc' not in keys:
            cmd_keys = [ key for key in valid_keys if 'cmd' in key]
            if len(cmd_keys) > 0:
                kwargs['rc'] = True

        self.logger.debug("Searching for client(s) meeting : {}".format(kwargs))
        r = requests.get(self.server+'/query', cert=self.cert, params=kwargs).json()
        results = {}
        if 'clients' in r:
            results['clients'] = [Client(c, profile=self.profile) for c in r['clients']]
            if 'commands' in r:
                results['commands'] = [Command(c, profile=self.profile) for c in r['commands']]
            if 'client_id_list' in r:
                results['client_id_list'] = r['client_id_list']
            return results
        return r

    def get_command(self, cid):
        """ Get a command by it's id.

        :param int cid: The id of a lerc command.
        :return: A lerc_api.Command object or False
        """
        r = self.query(cmd_id=cid, rc=True)
        # querying by command id should always return a single command
        if 'commands' in r and len(r['commands']) == 1:
            return r['commands'][0]
        self.logger.warning("No command result for {} : {}".format(cid, r))
        return False

    def get_host(self, hostname):
        """Query for a lerc by hostname. Return a Client object if one lerc is found, else a list of Clients.

        :param str hostname: The hostname of a lerc
        :return: False if no lerc by the hostname, a lerc_api.Client object if a sigle lerc, else a list of lerc_api.Clients
        """
        result = self.query(hostname=hostname)
        if 'clients' in result:
            if len(result['clients']) > 1:
                self.logger.error("More than one result for query by hostname")
                return result['client']
            elif len(result['clients']) == 1:
                return result['clients'][0]
            else:
                self.logger.info("No lerc found by this hostname: {}".format(hostname))
                return False

    def get_client(self, cid):
        """Get a client by it's id.

        :param int cid: The id of a lerc
        :return: A lerc_api.Client object of False
        """
        # should only ever return one lerc
        r = self.query(client_id=cid)
        if 'clients' in r and len(r['clients']) == 1:
            return r['clients'][0]
        self.logger.info("No Client with ID {} exists.".format(cid))
        return False

    def yield_hosts(self):
        """Yeild every lerc client the server knows about.
        """
        # The server will give us a list of valid client ids when we give a valid
        # query that returns no results -- a client by id zero does not exist.
        result = self.query(client_id=0)
        if 'client_id_list' not in result:
            self.logger.error("Unexpected result from lerc server : {}".format(result))
            return False
        client_ids = result['client_id_list']
        for id in client_ids:
            yield self.get_client(id)





class TablePrinter(object):
    """Print a list of dicts as a table.

    :param fmt: list of tuple(heading, key, width)
    :param sep: string, separation between columns
    :param ul: string, character to underline column label, or None for no underlining
    :return: A string representation of the table, ready to be printed

    Each tuple in the fmt list of tuples is like so:

    :heading: str, column label
    :key: dictionary key to value to print
    :width: int, column width in chars
    """
    def __init__(self, fmt, sep=' ', ul=None):
        super(TablePrinter,self).__init__()
        self.fmt   = str(sep).join('{lb}{0}:{1}{rb}'.format(key, width, lb='{', rb='}') for heading,key,width in fmt)
        self.head  = {key:heading for heading,key,width in fmt}
        self.ul    = {key:str(ul)*width for heading,key,width in fmt} if ul else None
        self.width = {key:width for heading,key,width in fmt}

    def row(self, data):
        return self.fmt.format(**{ k:str(data.get(k,''))[:w] for k,w in self.width.items() })

    def __call__(self, dataList):
        _r = self.row
        res = [_r(data) for data in dataList]
        res.insert(0, _r(self.head))
        if self.ul:
            res.insert(1, _r(self.ul))
        return '\n'.join(res)



#!/usr/bin/env python3

import os
import sys
import time
import argparse
import logging
import coloredlogs
import pprint
from lerc_control import lerc_api, collect
from lerc_control.scripted import execute_script
from lerc_control.helpers import TablePrinter

# configure logging #
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s - %(name)s - [%(levelname)s] %(message)s')
# set noise level
logging.getLogger('requests').setLevel(logging.WARNING)
logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
logging.getLogger('lerc_api').setLevel(logging.INFO)
logging.getLogger('lerc_control').setLevel(logging.INFO)
#logging.getLogger('lerc_control.scripted').setLevel(logging.INFO)
#logging.getLogger('lerc_control.collect').setLevel(logging.INFO)

logger = logging.getLogger('lerc_ui')
coloredlogs.install(level='INFO', logger=logger)

try:
    from lerc_control import deploy_lerc
except:
    pass

def main():

    parser = argparse.ArgumentParser(description="User interface to the LERC control server")
    # LERC environment choices
    config = lerc_api.load_config()
    env_choices = [ sec for sec in config.sections() if config.has_option(sec, 'server') ]
    parser.add_argument('-e', '--environment', action="store", default='default', 
                        help="specify an environment to work with. Default='default'", choices=env_choices)
    parser.add_argument('-d', '--debug', action="store_true", help="set logging to DEBUG")
    parser.add_argument('-c', '--check', action="store", help="check on a specific command id")
    parser.add_argument('-r', '--resume', action='store', help="resume a pending command id") 
    parser.add_argument('-g', '--get', action='store', help="get results for a command id")

    subparsers = parser.add_subparsers(dest='instruction') #title='subcommands', help='additional help')

    # Query
    parser_query = subparsers.add_parser('query', help="Query the LERC Server")
    parser_query.add_argument('query', help="The search you want to run. Enter 'fields' to see query fields.")
    parser_query.add_argument('-rc', '--return-commands', action='store_true', help="Return command results (even if no cmd fields specified)")
 
    # Initiate new LERC commands
    parser_run = subparsers.add_parser('run', help="Run a shell command on the host.")
    parser_run.add_argument('hostname', help="the host you'd like to work with")
    parser_run.add_argument('command', help='The shell command for the host to execute`')
    parser_run.add_argument('-a', '--async', action='store_true', help='Set asynchronous to true (do NOT wait for output or command to complete)')

    parser_upload = subparsers.add_parser('upload', help="Upload a file from the client to the server")
    parser_upload.add_argument('hostname', help="the host you'd like to work with")
    parser_upload.add_argument('file_path', help='the file path on the client')

    parser_download = subparsers.add_parser('download', help="Download a file from the server to the client")
    parser_download.add_argument('hostname', help="the host you'd like to work with")
    parser_download.add_argument('file_path', help='the path to the file on the server')
    parser_download.add_argument('-f', '--local-file', help='where the client should write the file')

    parser_quit = subparsers.add_parser('quit', help="tell the client to uninstall itself")
    parser_quit.add_argument('hostname', help="the host you'd like to work with")

    # response functions
    parser_collect = subparsers.add_parser('collect', help="Default (no arguments): perform a full lr.exe collection")
    parser_collect.add_argument('-d', '--directory', action='store', help="Compress contents of a client directory and collect")
    parser_collect.add_argument('hostname', help="the host you'd like to work with")

    parser_contain = subparsers.add_parser('contain', help="Contain an infected host")
    parser_contain.add_argument('hostname', help="the host you'd like to work with")
    parser_contain.add_argument('-on', action='store_true', help="turn on containment")
    parser_contain.add_argument('-off', action='store_true', help="turn off containment")
    parser_contain.add_argument('-s', '--status', action='store_true', help="Get containment status of host")

    parser_script = subparsers.add_parser('script', help="run a scripted routine on this lerc.")
    parser_script.add_argument('hostname', help="the host you'd like to work with")
    parser_script.add_argument('-l', '--list-scripts', action='store_true', help="list scripts availble to lerc_ui")
    parser_script.add_argument('-s', '--script-name', help="provide the name of a script to run")
    parser_script.add_argument('-f', '--file-path', help="the path to a custom script you want to execute")

    args = parser.parse_args()

    if args.debug:
        logging.getLogger('lerc_api').setLevel(logging.DEBUG)
        logging.getLogger('lerc_control').setLevel(logging.DEBUG)
        coloredlogs.install(level='DEBUG', logger=logger)

    # a local lerc_session will be needed to go any further
    ls = lerc_api.lerc_session(profile=args.environment)

    if args.instruction == 'query':
        if args.query == 'fields':
            print("\nAvailable query fields:\n")
            fmt = [ ('Field', 'field', 14),
                    ('Description', 'description', 80) ]
            print( TablePrinter(fmt, sep='  ', ul='=')(lerc_api.QUERY_FIELD_DESCRIPTIONS) )
            print()
            print("NOTE:")
            print("  1) Fields are ANDed by default. Fields can be negated by appending '-' or '!' to the front of the field (no space) or by specifying 'NOT ' in front of the field (space).")
            print("  2) A leading '-' with no space in front will cause the argument parser to misinterpret the query as a command line argument option.")
            print() 
            sys.exit()
        query = lerc_api.parse_lerc_server_query(args.query)
        if args.return_commands: 
            query['rc'] = True
        results = ls.query(**query)
        clients = [ c.get_dict for c in results['clients']]
        fmt = [ ('ID', 'id', 5),
                ('Hostname', 'hostname', 20),
                ('Status', 'status', 11),
                ('Version', 'version', 8),
                ('Sleep Cycle', 'sleep_cycle', 11),
                ('Install Date', 'install_date', 20),
                ('Last Activity', 'last_activity', 20),
                ('Company ID', 'company_id', 10)]
        print("\nClient Results:\n")
        print( TablePrinter(fmt, sep='  ', ul='=')(clients))
        print("Total Client Results:{}".format(len(clients)))
        print()
        commands = [ c.get_dict for c in results['commands']]
        if commands:
            fmt = [ ('ID', 'command_id', 9),
                    ('Client ID', 'client_id', 9),
                    ('Hostname', 'hostname', 20),
                    ('Operation', 'operation', 11),
                    ('Status', 'status', 9)]
                  #  ('Version', 'version', 8),
                  #  ('Sleep Cycle', 'sleep_cycle', 11),
                  #  ('Install Date', 'install_date', 20),
                  #  ('Last Activity', 'last_activity', 20)]
            print("\nCommand Results:\n")
            print( TablePrinter(fmt, sep='  ', ul='=')(commands))
            #for cmd in commands:
            #    print(cmd)
            print() 
        sys.exit()

    # root options
    if args.check:
        command = ls.get_command(args.check)
        print(command)
        if command.status == 'ERROR':
            print("ERROR Report:")
            pprint.pprint(command.get_error_report(), indent=5)
        sys.exit()
    elif args.get:
        command = ls.get_command(args.get)
        if command:
            command.get_results(chunk_size=16384)
            print(command)
        sys.exit()
    elif args.resume:
        command = ls.get_command(args.resume)
        command.wait_for_completion()
        if command:
            print(command)
        sys.exit()

    # if we're here, then an instructions been specified and the args.hostname is a thing
    client = ls.get_host(args.hostname)
    if isinstance(client, list):
        logger.critical("More than one result. Not handled yet..")
        for c in client:
            print(c)
            print()
        sys.exit(1) 

    # Auto-Deployment Jazz
    bad_status = False
    if client:
        if client.status == 'UNINSTALLED' or client.status == 'UNKNOWN':
            logger.info("Non-working client status : {}".format(client.status))
            bad_status = True

    if not client or bad_status:
        config = ls.get_config
        if config.has_option('default', 'cb_auto_deploy') and not config['default'].getboolean('cb_auto_deploy'):
            logger.info("CarbonBlack auto-deployment turned off. Exiting..")
            sys.exit(0)
        logger.info("Attempting to deploy lerc with CarbonBlack..")
        try:
            from cbapi import auth
            from deploy_lerc import deploy_lerc, CbSensor_search
        except:
            logger.error("Failed to import deployment functions from lerc_control.deploy_lerc OR cbapi.")
            sys.exit(1)
        logging.getLogger('lerc_control.deploy_lerc').setLevel(logging.ERROR)
        environments = auth.CredentialStore("response").get_profiles()
        sensors = []
        logger.debug("Trying to find the sensor in the available carbonblack environments.")
        for env in environments:         
            sensor = CbSensor_search(env, args.hostname)
            if sensor:
                logger.debug("Found sensor in {} environment".format(env))
                sensors.append((env, sensor))
        if len(sensors) > 1:
            logger.error("A CarbonBlack Sensor was found by that hostname in multiple environments.")
            sys.exit(1)
        elif sensors:
            logging.getLogger('lerc_control.deploy_lerc').setLevel(logging.INFO)
            sensor = sensors[0][1]
            config = lerc_api.check_config(config, required_keys=['lerc_install_cmd', 'client_installer'])
            result = deploy_lerc(sensor, config[sensors[0][0]]['lerc_install_cmd'], lerc_installer_path=config['default']['client_installer'])
            if result: # modify deploy_lerc to use new client objects
                logger.info("Successfully deployed lerc to this host: {}".format(args.hostname))
                client = ls.get_host(args.hostname)
        else:
            logger.error("Didn't find a sensor in CarbonBlack by this hostname")
            sys.exit(0)
 
    # collections
    profile=args.environment if args.environment else 'default'
    if args.instruction == 'collect':
        if not args.debug:
            logging.getLogger('lerc_api').setLevel(logging.WARNING)
        if args.directory:
            commands = collect.get_directory(client, args.directory)
        else:
            collect.full_collection(client)
        sys.exit(0)

    if args.instruction == 'script':
        config = ls.get_config
        if args.list_scripts:
            if not config.has_section('scripts'):
                print("\nNo pre-existing scripts have been made availble.")
                sys.exit(0)           
            print("\nAvailable scripts:")
            for sname in config['scripts']:
                print("\t{}".format(sname))
            print()
            sys.exit(0)
        elif args.script_name:
            if not config.has_option('scripts', args.script_name):
                logger.error("{} is not a defined script".format(args.script_name))
                sys.exit(1)
            script_path = config['scripts'][args.script_name]
            commands = execute_script(client, script_path)
            sys.exit(0)
        elif args.file_path:
            if not os.path.exists(args.file_path):
                logger.error("Could not find script file at '{}'".format(args.file_path))
                sys.exit(1)
            commands = execute_script(client, args.file_path)
            sys.exit(0)
        else:
            logger.info("No argument was specified for the script command. Exiting.")
            sys.exit(0)

    # Else, see if we're running a command directly
    cmd = None
    if args.instruction == 'run':
        if args.async:
            cmd = client.Run(args.command, async=args.async)
        else:
            cmd = client.Run(args.command)

    elif args.instruction == 'contain':
        if args.on:
            client.contain()
        elif args.off:
            client.release_containment()
        elif args.status:
            print("Containment status check not yet implemented.")

    elif args.instruction == 'download':
        # if client_file_path is not specified the client will write the file to it's local dir
        analyst_file_path = os.path.abspath(args.file_path)
        file_name = args.file_path[args.file_path.rfind('/')+1:]
        if args.local_file is None:
            args.local_file = file_name
        cmd = client.Download(file_name, client_file_path=args.local_file, analyst_file_path=analyst_file_path)

    elif args.instruction == 'upload':
        cmd = client.Upload(args.file_path)

    elif args.instruction == 'quit':
        cmd = client.Quit()
    elif args.check:
        command = ls.get_command(args.check)
        print(command)
        sys.exit()
    elif args.get:
        command = ls.get_command(args.get)
        if command:
            command.get_results(chunk_size=16384)
            print(command)
        sys.exit()
    elif args.resume:
        command = ls.get_command(args.resume)
        command.wait_for_completion()
        if command:
            print(command)
        sys.exit()
    else:
        print(client)
        sys.exit()

    if not cmd:
        sys.exit(1)

    start_time = time.time() 
    if not cmd.wait_for_completion():
        logger.warning("{} (ID:{}) command went to a {} state. Exiting.".format(cmd.operation, cmd.id, cmd.status))
        sys.exit(1)
    logger.info("{} command {} completed successfully".format(cmd.operation, cmd.id))
    cmd.get_results()

    print(cmd)

    sys.exit()





import os
import time
import subprocess
import shlex
import pprint
import logging

from lerc_control import lerc_api

logger = logging.getLogger("lerc_control."+__name__)

# Get the working lerc_control directory
BASE_DIR = os.path.dirname(os.path.abspath(__file__))

def get_directory(lerc, dir_path):
    """Compress a directory with 7zip and upload the compressed result.

    :param lerc_api.Client lerc: A lerc_api.Client
    :param str dir_path: The path to the client directory
    :return: list of lerc_api.Command objects on success, else False
    """
    if not isinstance(lerc, lerc_api.Client):
        logger.error("Argument is of type:{} instead of type lerc_api.Client".format(type(lerc)))
        return False

    config = lerc._ls.get_config
    required_keys=['7za_path', '7za_dir_cmd', 'client_working_dir']
    if not lerc_api.check_config(config, required_keys=required_keys):
        logger.error("^ Missing required config items")
        return False
    profile = lerc._ls.profile
    section = profile + "_collect"
    _7za_path = config[section]['7za_path']
    _7za_cmd = config[section]['7za_dir_cmd']
    default_client_dir = config[profile]['client_working_dir']

    if not os.path.exists(_7za_path):
        if _7za_path[0] == '/':
            _7za_path = BASE_DIR + _7za_path
        else:
            _7za_path = BASE_DIR + '/' + _7za_path

    if not os.path.exists(_7za_path):
        logger.error("'{}' does not exist".format(_7za_path))
        return False

    commands = []
    cmd = lerc.Download(_7za_path)
    logger.info("Issued CID={} to download 7za.exe.".format(cmd.id))
    commands.append(cmd)

    cmd = lerc.Run(_7za_cmd.format(lerc.hostname+'_dirOfInterest' , dir_path))
    logger.info("Issued CID={} to run 7za on '{}'".format(cmd.id, dir_path))
    commands.append(cmd)

    outputfile = default_client_dir + '{}_dirOfInterest.7z'.format(lerc.hostname)
    upCmd = lerc.Upload(outputfile)
    logger.info("Issued CID={} to upload {}_dirOfInterest.7z".format(upCmd.id, lerc.hostname))
    logger.info("Waiting for the upload command to reach completion ... ")
    commands.append(upCmd.wait_for_completion())

    cmd = lerc.Run('Del "{}" && Del 7za.exe'.format(outputfile))
    logger.info("Issued CID={} to to delete '{}' and 7za.exe".format(cmd.id, lerc.hostname))
    commands.append(cmd)

    logger.info("Getting result from the control server..".format(lerc.hostname))
    if upCmd.get_results(file_path='{}_dirOfInterest.7z'.format(lerc.hostname)):
        logger.info("Wrote {}_dirOfInterest.7z".format(lerc.hostname))
    return commands


def full_collection(lerc):
    #########################################################################################
    ### This is an Integral Defense custom module built for a private collection package. ###
    ### :param lerc_api.Client lerc: A lerc Client object                                 ###
    #########################################################################################
    if not isinstance(lerc, lerc_api.Client):
        logger.error("Argument is of type:{} instead of type lerc_api.Client".format(type(lerc)))
        return False

    # Config config items exist and get them
    config = lerc._ls.get_config
    required_keys = ['lr_path', 'extract_cmd', 'collect_cmd', 'output_dir', 'streamline_path', 'client_working_dir']
    if not lerc_api.check_config(config, required_keys=required_keys):
        logger.error("^ Missing required configuration item(s)")
        return False
    profile = lerc._ls.profile
    collect_profile = profile+"_collect"
    client_workdir = config[profile]['client_working_dir']
    lr_path = config[collect_profile]['lr_path']
    extract_cmd = config[collect_profile]['extract_cmd']
    collect_cmd = config[collect_profile]['collect_cmd']
    output_dir = config[collect_profile]['output_dir']
    streamline_path = config[collect_profile]['streamline_path']

    commands = []

    logger.info("Starting full Live Response collection on {}.".format(lerc.hostname))

    # for contriving the output filename
    local_date_str_cmd = lerc.Run('date /t')
    # Delete any existing LR artifacts
    lerc.Run("DEL /S /F /Q lr && rmdir /S /Q lr")
    # download the package
    lr_download = lerc.Download(lr_path)
    logger.info("Issued CID={} for client to download {}.".format(lr_download.id, lr_path))
    # extract the package
    result = lerc.Run(extract_cmd)
    logger.info("Issued CID={} to extract lr.exe on the host.".format(result.id))
    # run the collection
    collect_command = lerc.Run(collect_cmd)
    logger.info("Issued CID={} to run {}.".format(collect_command.id, collect_cmd))
    # finish contriving the output filename
    output_filename = None
    local_date_str_cmd.refresh()
    while True:
        if local_date_str_cmd.status == 'COMPLETE':
            dateStr = local_date_str_cmd.get_results(return_content=True).decode('utf-8')
            logger.debug("Got date string of '{}'".format(dateStr))
            # Mon 11/19/2018 -> 20181119      
            dateStr = dateStr.split(' ')[1].split('/')
            dateStr =  dateStr[2]+dateStr[0]+dateStr[1]
            # hostname.upper() because streamline.py expects uppercase
            output_filename = lerc.hostname.upper() + "." + dateStr + ".7z"
            break
        # wait five seconds before asking the server again
        time.sleep(5)
        local_date_str_cmd.refresh() 
    # collect the output file
    upload_command = lerc.Upload(client_workdir + output_dir + output_filename)
    logger.info("Issued CID={} to upload output at: '{}'".format(upload_command.id, client_workdir + output_dir + output_filename))
    # Stream back collect.bat output as it comes in
    logger.info("Streaming collect.bat output ... ")
    position = 0
    while True:
        collect_command.refresh()
        if collect_command.status == 'STARTED':
            if collect_command.filesize > 0:
                results = collect_command.get_results(return_content=True, position=position)
                if len(results) > 0:
                    position += len(results)
                    print(results.decode('utf-8'))
            time.sleep(1)
        elif collect_command.status == 'COMPLETE':
            if position < collect_command.filesize:
                results = collect_command.get_results(return_content=True, position=position)
                if len(results) > 0:
                    position += len(results)
                    print(results.decode('utf-8'))
            elif position >= collect_command.filesize:
                break
        elif collect_command.status == 'UNKNOWN' or collect_command.status == 'ERROR':
            logger.error("Collect command went to {} state : {}".format(collect_command.status, collect_command))
            return False
        time.sleep(5)
    logger.info("Waiting for '{}' upload to complete.".format(output_filename))
    upload_command.wait_for_completion()
    #commands.append(upload_command)
    if upload_command.status == 'COMPLETE':
        logger.info("Upload command complete. Telling lerc to delete the output file on the client")
        commands.append(lerc.Run('DEL /S /F /Q "{}"'.format(client_workdir + output_dir + output_filename)))

    # finally, stream the collection from the server to the cwd
    logger.info("Streaming {} from server..".format(output_filename))
    upload_command.get_results(file_path=output_filename)
    # Call steamline on the 7z lr package
    logger.info("[+] Starting streamline on {}".format(output_filename))
    args = shlex.split(streamline_path + " " + output_filename)
    try:
        subprocess.Popen(args).wait()
        logger.info("[+] Streamline complete")
    except Exception as e:
        logger.error("Exception with Streamline: {}".format(e))

    return True



#!/usr/bin/env python3
#/data/home/carbonblack/env3/bin/python3

import os
import re
import sys
import time
import argparse
import datetime
import json
import pprint
import coloredlogs

from dateutil import tz
from configparser import ConfigParser

from lerc_control import lerc_api
import logging

# configure logging #
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s - %(name)s - [%(levelname)s] %(message)s')
# set noise level
logging.getLogger('requests').setLevel(logging.WARNING)
logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
logging.getLogger('lerc_api').setLevel(logging.INFO)
logging.getLogger('cbapi').setLevel(logging.WARNING)

logger = logging.getLogger("lerc_control."+__name__)
coloredlogs.install(level='DEBUG', logger=logger)


try:
    from cbapi import auth
    from cbapi.response import *
    from cbapi.errors import ApiError, ObjectNotFoundError, TimeoutError
except:
    pass
    #logging.debug("cbapi is not installed")
    sys.exit(1)

HOME_DIR = os.path.dirname(os.path.realpath(__file__)) 


def eastern_time(timestamp):
    eastern_timebase = tz.gettz('America/New_York')
    eastern_time = timestamp.replace(tzinfo=tz.gettz('UTC'))
    return eastern_time.astimezone(eastern_timebase).strftime('%Y-%m-%d %H:%M:%S.%f%z')


def go_live(sensor):
    start_time = time.time()
    timeout = 604800 # seven days
    current_day = 0
    lr_session = None
    while time.time() - start_time < timeout:
        try:
            lr_session = sensor.lr_session()
            logger.info("LR session started at {}".format(time.ctime()))
            break
        except TimeoutError:
            elapsed_time = time.time() - start_time
            if current_day != elapsed_time // 86400:
                current_day+=1
                logger.info("24 hours of timeout when polling for LR session")
                logger.info("Attempting LR session again on {} @ {}".format(args.sensor,
                                                                        time.ctime()))
    return lr_session


def deploy_lerc(sensor, install_cmd, lerc_installer_path=None):

    if not isinstance(sensor, models.Sensor):
        logger.error("Cb models.Sensor object required.")
        return False

    hostname = sensor.hostname
    default_lerc_path = '/opt/lerc_control/lercSetup.msi'

    if lerc_installer_path is None:
        config = lerc_api.load_config('default', required_keys=['client_installer'])
        lerc_installer_path = config['default']['client_installer']

    # create lerc session
    ls = lerc_api.lerc_session()
    # check and see if the client's already installed
    client = None
    try:
        client = ls.get_host(hostname)
    except:
        logger.warning("Can't reach the lerc control server")

    previously_installed = proceed_with_force = None
    if client:
        if client.status != 'UNINSTALLED':
            errmsg = "lerc server reports the client is already installed on a system with this hostname:\n{}"
            errmsg = errmsg.format(client)
            logger.warning(errmsg)
            proceed_with_force = input("Proceed with fresh install? (y/n) [n] ") or 'n'
            proceed_with_force = True if proceed_with_force == 'y' else False
            if not proceed_with_force:
                return None
        else:
            previously_installed = True
            logger.info("A client was previously uninstalled on this host: {}".format(client))

    lr_session = None
    try:
        logger.info(".. attempting to go live on the host with CarbonBlack..")
        lr_session = go_live(sensor)
    except Exception as e:
        logger.error("Failed to start Cb live response session on {}".format(hostname))
        return False

    with lr_session:

        if proceed_with_force:
            uninstall_cmd = "msiexec /x C:\Windows\Carbonblack\lercSetup.msi /quiet /qn /norestart /log C:\Windows\Carbonblack\lerc_Un-Install.log"
            logger.info("~ checking for installed lerc client on host..")
            result = lr_session.create_process("sc query lerc", wait_timeout=60, wait_for_output=True)
            result = result.decode('utf-8')
            logger.info("~ Got service query result:\n{}".format(result))
            if 'SERVICE_NAME' in result and 'lerc' in result:
                logger.info("~ attempting to uninstall lerc..")
                result = lr_session.create_process(uninstall_cmd, wait_timeout=60, wait_for_output=True)
                logger.info("~ Post uninstall service query:\n {}".format(lr_session.create_process("sc query lerc",
                                                                            wait_timeout=60, wait_for_output=True).decode('utf-8')))

        logger.info("~ dropping current Live Endpoint Response Client msi onto {}".format(hostname))
        logger.debug("Dropping this msi : {}".format(lerc_installer_path))
        filedata = None
        with open(lerc_installer_path, 'rb') as f:
            filedata = f.read()
        try:
            lr_session.put_file(filedata, "C:\\Windows\\Carbonblack\\lercSetup.msi")
        except Exception as e:
            if 'ERROR_FILE_EXISTS' in str(e):
                logger.info("~ lercSetup.msi already on host. Deleting..")
                lr_session.delete_file("C:\\Windows\\Carbonblack\\lercSetup.msi")
                lr_session.put_file(filedata, "C:\\Windows\\Carbonblack\\lercSetup.msi")
                #pass
            else:
                raise e

        logger.info("~ installing the lerc service")
        result = lr_session.create_process(install_cmd, wait_timeout=60, wait_for_output=True)

    def _get_install_log(logfile=None):
        logger.info("Getting install log..")
        logfile = logfile if logfile else r"C:\\Windows\\Carbonblack\\lerc_install.log"
        content = lr_session.get_file(logfile)
        with open(hostname+"_lerc_install.log", 'wb') as f:
            f.write(content)
        logger.info("wrote log file to {}_lerc_install.log".format(hostname))


    wait = 5 #seconds
    attempts = 6
    if previously_installed:
        attempts += attempts
    logger.info("~ Giving client up to {} seconds to check in with the lerc control server..".format(attempts*wait))

    for i in range(attempts):
        try:
            client = ls.get_host(hostname)
        except:
            logger.warning("Can't reach the lerc control server")
            break
        if client:
            if client.status != 'UNINSTALLED':
                break
        logger.info("~ giving the client {} more seconds".format(attempts*wait - wait*i))
        time.sleep(wait)

    if not client:
        logger.warning("failed to auto-confirm install with lerc server.")
        _get_install_log()
        return None
    elif previously_installed and client.status == 'UNINSTALLED':
        logger.warning("Failed to auto-confirm install. Client hasn't checked in.")
        _get_install_log()
        return False

    logger.info("Client installed on {} at '{}' - status={} - last check-in='{}'".format(hostname,
                                 client.install_date, client.status, client.last_activity))
    return client


#Get the right CarbonBlack sensor
def CbSensor_search(profile, hostname):
    cb = CbResponseAPI(profile=profile)
    sensor = None
    try:
        logger.debug("Getting the sensor object from carbonblack")
        return cb.select(Sensor).where("hostname:{}".format(hostname)).one()
    except TypeError as e:
        # Appears to be bug in cbapi library here -> site-packages/cbapi/query.py", line 34, in one
        # Raise MoreThanOneResultError(message="0 results for query {0:s}".format(self._query))
        # That raises a TypeError 
        if 'non-empty format string passed to object' in str(e):
            try: # accounting for what appears to be an error in cbapi error handling
                result = cb.select(Sensor).where("hostname:{}".format(hostname))
                if isinstance(result[0], models.Sensor):
                    print()
                    logger.warn("MoreThanOneResult Error searching for {0:s}".format(hostname))
                    print("\nResult breakdown:")
                    sensor_ids = []
                    for s in result:
                        sensor_ids.append(int(s.id))
                        if int(s.id) == max(sensor_ids):
                            sensor = s
                        print()
                        print("Sensor object - {}".format(s.webui_link))
                        print("-------------------------------------------------------------------------------\n")
                        print("\tos_environment_display_string: {}".format(s.os_environment_display_string))
                        print()
                        print("\tstatus: {}".format(s.status))
                        print("\tsensor_id: {}".format(s.id))
                        print("\tlast_checkin_time: {}".format(s.last_checkin_time))
                        print("\tnext_checkin_time: {}".format(s.next_checkin_time))
                        print("\tsensor_health_message: {}".format(s.sensor_health_message))
                        print("\tsensor_health_status: {}".format(s.sensor_health_status))
                        print("\tnetwork_interfaces:")
                    print()
                    default_sid = max(sensor_ids)
                    choice_string = "Which sensor do you want to use?\n"
                    for sid in sensor_ids:
                        choice_string += "\t- {}\n".format(sid)
                    choice_string += "\nEnter one of the sensor ids above. Default: [{}]".format(default_sid)
                    user_choice = int(input(choice_string) or default_sid)
                    for s in result:
                        if user_choice == int(s.id):
                            return s
            except Exception as e:
                if sensor is None:
                    logger.warning("A sensor by hostname '{}' wasn't found in this environment".format(hostname))
                    return False
                logger.error("{}".format(str(e)))
                return False
    except Exception as e:
        logger.error("{}".format(str(e)))
        return False


def main(argv):

    parser = argparse.ArgumentParser(description="put file on CB sensor")
    parser.add_argument('company', choices=auth.CredentialStore("response").get_profiles(),
                        help='specify an environment you want to work with.')

    parser.add_argument('hostname', help="the name of the host to deploy the client to")
    parser.add_argument('-p', '--package', help="the msi lerc package to install")
    args = parser.parse_args()

    print(time.ctime() + "... starting")

    # ignore the proxy
    #del os.environ['https_proxy']

    # get the config items we need
    required_keys = ['client_installer', 'lerc_install_cmd']
    config = lerc_api.load_config(required_keys=required_keys)
    default_lerc_path = config['default']['client_installer'] #'/opt/lerc/lercSetup.msi'

    sensor = CbSensor_search(args.company, args.hostname)

    result = deploy_lerc(sensor, config[args.company]['lerc_install_cmd'], lerc_installer_path=args.package)
    if result:
        print()
        print(result)
        print()



if __name__ == "__main__":
    result = main(sys.argv[1:])
    if result != 1:
        print(time.ctime() + "...Done.")
    sys.exit(result)




import os
import logging
import pprint
from lerc_control import lerc_api
from configparser import ConfigParser

logger = logging.getLogger("lerc_control."+__name__)

REQUIRED_CMD_KEYS = ['operation']
OPTIONAL_CMD_KEYS = ['wait_for_completion', 'get_results']

REQUIRED_OP_KEY_MAP = {'RUN': ['command'],
                       'UPLOAD': ['path'],
                       'DOWNLOAD': ['file_path'],
                       'QUIT': []}
OPTIONAL_OP_KEY_MAP = {'RUN': ['async_run', 'write_results_path'],
                       'UPLOAD': ['write_results_path'],
                       'DOWNLOAD': ['client_file_path'],
                       'QUIT': []}

# Get the working lerc_control directory
BASE_DIR = os.path.dirname(os.path.abspath(__file__))

def script_missing_required_keys(config, KEYS):
    for key in KEYS:
        for section in config.sections():
            if not config.has_option(section, key):
                logger.error("{} is missing required key: {}".format(section, key))
                return True
    return False

def operation_missing_required_keys(config, section, KEYS):    
    for key in KEYS:
        if not config.has_option(section, key):
            logger.error("{} is missing required operation key:{} for operation:{}".format(section, key, config[section]['operation']))
            return True
    return False

def execute_script(lerc, script_path):
    """Execute a script on this host.

    :param lerc_api.Client lerc: A lerc_api.Client object.
    :param str script_path: the path to the script
    :return: a dictionary of the commands issued
    """

    if not isinstance(lerc, lerc_api.Client):
        logger.error("Argument is of type:{} instead of type lerc_api.Client".format(type(lerc)))
        return False

    config = lerc._ls.get_config
    profile = lerc._ls.profile
    default_client_dir = config[profile]['client_working_dir']

    script = ConfigParser()
    if not os.path.exists(script_path):
        if script_path[0] == '/':
            script_path = BASE_DIR + script_path
        else:
            script_path = BASE_DIR + '/' + script_path
    script.read(script_path)
 
    if script_missing_required_keys(script, REQUIRED_CMD_KEYS):
        return False

    command_history = {}

    script_name = script_path[script_path.rfind('/')+1:script_path.rfind('.')]
    # make sure requirements are met first
    for command in script.sections():
        op =  script[command]['operation'].upper()
        if op not in REQUIRED_OP_KEY_MAP:
            logger.error("{} is not a recognized lerc operation!".format(op))
            return False
        if operation_missing_required_keys(script, command, REQUIRED_OP_KEY_MAP[op]):
            return False

    logger.info("Beginning execution of {}".format(script_name))
    for command in script.sections():
        logger.info("Processing {}".format(command))
        command_history[command] = {}
        op =  script[command]['operation'].upper()

        get_results = False
        if 'get_results' in script[command]:
            get_results = script[command].getboolean('get_results')
        # should only ever be in run and upload commands
        write_results_path = None
        if 'write_results_path' in script[command]:
            write_results_path = script[command]['write_results_path']

        if op == 'RUN':
            async_run = False
            if 'async_run' in script[command]:
                async_run = script[command].getboolean('async_run')
            run_string = script[command]['command']
            cmd = lerc.Run(run_string, async=async_run)
            command_history[command] = cmd
            command_history[command].get_the_results = get_results
            command_history[command].write_results_path = write_results_path
            logger.info("Issued : Run - CID={} - {}".format(cmd.id, run_string))
        elif op == 'DOWNLOAD':
            client_file_path = None
            if 'client_file_path' in script[command]:
                client_file_path = script[command]['client_file_path']
            file_path = script[command]['file_path']
            if not os.path.exists(file_path):
                old_fp = file_path
                if file_path[0] == '/':
                    file_path = BASE_DIR + file_path
                else:
                    file_path = BASE_DIR + '/' + file_path
                if not os.path.exists(file_path):
                    logger.error("Not found: '{}' OR '{}'".format(old_fp, file_path))
                    return False
            cmd = lerc.Download(file_path, client_file_path=client_file_path)
            command_history[command] = cmd
            logger.info("Issued : Download - CID={} - {}".format(cmd.id, file_path))
        elif op == 'UPLOAD':
            path = script[command]['path']
            # if the script doesn't specify the full path, add default client working dir
            if '\\' not in path:
                path = default_client_dir + path
            cmd = lerc.Upload(path)
            command_history[command] = cmd
            command_history[command].get_the_results = get_results
            command_history[command].write_results_path = write_results_path
            logger.info("Issued : Upload - CID={} - {}".format(cmd.id, path))
        elif op == 'QUIT':
            cmd = lerc.Quit()
            command_history[command] = cmd
            logger.info("Issued : Quit - CID={}".format(cmd.id, path))

    logger.info("Checking to see if results need to be obtained ...")
    for command in command_history:
        cmd = command_history[command]
        if hasattr(cmd, 'get_the_results') and cmd.get_the_results:
            logger.info("Waiting for command {} to complete..".format(cmd.id))
            cmd.wait_for_completion()
            logger.info("Getting the results for command {}".format(cmd.id))
            cmd.get_results(file_path=cmd.write_results_path)

    return command_history
